
from .turtle_class import Turtle
from .canvas import Canvas
from .logging_canvas import LoggingCanvas
from .forwarding_canvas import ForwardingCanvas
from .pillow_canvas import PillowCanvas
from .tk_canvas import TkCanvas




from abc import abstractmethod
from math import pi

from .canvas import Canvas

class AbstractLoggingCanvas(Canvas):

    def __init__(self, width, height):
        super().__init__(width, height)
        self.drawn_turtle = None

    @abstractmethod
    def on_action(self, log_line):
        pass

    def draw_rectangular_line(self, start, end, color, width):
        self.on_action(['draw_rectangular_line', [start.x, start.y, end.x, end.y], color, width])

    def draw_circle(self, center, radius, color, width, is_filled, start, end):
        self.on_action(['draw_circle', [center.x, center.y, radius], color, width, is_filled, start, end])

    def fill_path(self, path, color):
        self.on_action(['fill_path', [edge.json_friendly for edge in path], color])

    def axis_aligned_rectangle(self, bottom_left, width, height, color):
        self.on_action(['axis_aligned_rectangle', [bottom_left.x, bottom_left.y], width, height, color])

    def set_bgcolor(self, color):
        self.on_action(['set_bgcolor', color])

    def clear(self):
        self.on_action(['clear'])

    def refreshed_turtle(self, drawn_turtle):
        self.on_action([
            'refreshed_turtle',
            drawn_turtle.json_friendly if drawn_turtle is not None else None
        ])
        self.drawn_turtle = drawn_turtle

    def set_speed(self, speed):
        self.on_action(['set_speed', speed])

    def exit_on_click(self):
        self.on_action(['exit_on_click'])

class LoggingCanvas(AbstractLoggingCanvas):
    def __init__(self, width, height):
        super().__init__(width, height)
        self.log = []
    def on_action(self, log_line):
        self.log.append(log_line)




from time import sleep
import unittest

from abstract_turtle import Turtle
from abstract_turtle.tk_canvas import TkCanvas
from .image_tests import PillowTest

MANUAL_INSPECT = False

class TkSmokeTest(PillowTest):
    def setUp(self, width=1000, height=1000):
        self.canvas = TkCanvas(width, height)
        self.turtle = Turtle(self.canvas)
        self.turtle.clear()
        self.turtle.bgcolor("white")
        if not MANUAL_INSPECT:
            import turtle
            turtle.speed(0)

    def assertImageEquals(self, path, hide_turtle=True):
        if MANUAL_INSPECT:
             # give a second to inspect the image
            sleep(1)

del PillowTest




from .forwarding_canvas import ForwardingCanvas
from .logging_canvas import LoggingCanvas
from .turtle_class import Turtle

__canvas = ForwardingCanvas(LoggingCanvas(1000, 1000))
__turtle = Turtle(__canvas)

for method_name in dir(__turtle):
    method = getattr(__turtle, method_name)
    if getattr(method, "is_turtle_method", False):
        globals()[method_name] = method

def set_canvas(canvas):
    __canvas.set_canvas(canvas)




import unittest

import numpy as np
from PIL import Image

from abstract_turtle import Turtle, PillowCanvas, ForwardingCanvas

# set this to true to create the images
REFRESH_IMAGES = False

class PillowTest(unittest.TestCase):

    def setUp(self, width=1000, height=1000):
        self.canvas = PillowCanvas(width, height)
        self.turtle = Turtle(self.canvas)

    def assertImageEquals(self, path, hide_turtle=True):
        if hide_turtle:
            self.turtle.hideturtle()
        data = self.canvas.export()
        if REFRESH_IMAGES:
            data.save(path, "png")
        else:
            saved_data = np.array(Image.open(path).convert("RGBA"))
            self.assertTrue(np.array_equal(data, saved_data))

    def test_square(self):
        t = self.turtle
        for _ in range(9):
            for _ in range(4):
                t.fd(100)
                t.rt(90)
            t.rt(40)
        self.assertImageEquals("test_img/squareflower.png")

    def test_centeredness(self):
        t = self.turtle
        t.setpos(0, 100)
        t.setpos(0, -100)
        t.setpos(0, 0)
        t.setpos(100, 0)
        t.setpos(-100, 0)
        t.setpos(0, 0)
        t.width(20)
        t.setpos(0, 0)
        t.rt(45)
        t.fd(100)
        self.assertImageEquals("test_img/centeredness.png")

    def test_multisegment_forward(self):
        t = self.turtle
        t.width(20)
        for _ in range(3):
            t.fd(100)
        self.assertImageEquals("test_img/multiseg_forward.png")

    def test_empty_star(self):
        t = self.turtle
        t.lt(18)
        for _ in range(5):
            t.fd(100)
            t.lt(180 - 36)
            t.fd(100)
            t.rt(72)
        self.assertImageEquals("test_img/empty_star.png")

    def test_full_star(self):
        t = self.turtle
        t.fillcolor(255, 0, 0)
        t.width(10)
        t.begin_fill()
        t.lt(18)
        for _ in range(5):
            t.fd(100)
            t.lt(180 - 36)
            t.fd(100)
            t.rt(72)
        t.end_fill()
        self.assertImageEquals("test_img/full_star.png")

    def test_circle(self):
        t = self.turtle
        t.fd(100)
        t.circle(100)
        t.fd(100)
        self.assertImageEquals("test_img/basic_circle.png")

    def test_thick_circle(self):
        t = self.turtle
        t.width(10)
        t.circle(100)
        t.fd(100)
        self.assertImageEquals("test_img/thick_circle.png")

    def test_background(self):
        t = self.turtle
        t.pencolor(255, 255, 255)
        t.fd(100)
        t.bgcolor(255, 255, 0)
        t.rt(90)
        t.fd(100)
        self.assertImageEquals("test_img/background_test.png")

    def test_multicolor(self):
        t = self.turtle
        t.pencolor(255, 255, 0)
        t.fd(100)
        t.pencolor(0, 255, 255)
        t.fd(100)
        t.pencolor(0, 255, 0)
        t.fd(100)
        self.assertImageEquals("test_img/multicolor.png")

    def test_multi_directions(self):
        t = self.turtle
        for _ in range(8):
            t.fd(20)
            t.pu()
            t.fd(10)
            t.pd()
            t.rt(45)
        self.assertImageEquals("test_img/multi_directions.png")

    def test_setvarious_things(self):
        t = self.turtle
        t.setx(100)
        t.sety(100)
        self.assertEqual((100, 100), t.pos())
        t.setx(0)
        t.sety(0)
        t.setheading(-45)
        t.fd(100)
        self.assertImageEquals("test_img/set_various_things.png")

    def test_dot(self):
        t = self.turtle
        t.width(50)
        t.dot()
        t.pencolor(0, 0, 255)
        t.dot(50)

        t.width(1)
        t.pencolor(255, 0, 0)
        t.fd(50)
        t.pencolor(0, 255, 0)
        t.fd(50)
        self.assertImageEquals("test_img/dot.png")

    def test_backward(self):
        t = self.turtle
        t.fd(100)
        t.rt(45)
        t.bk(100)
        self.assertImageEquals("test_img/bk.png")

    def test_clear(self):
        t = self.turtle
        t.fd(100)
        t.rt(45)
        t.bgcolor(255, 0, 0)
        t.clear()
        t.rt(90)
        t.fd(100)
        self.assertImageEquals("test_img/clear.png")

    def test_reset(self):
        t = self.turtle
        t.fd(100)
        t.rt(45)
        t.bgcolor(255, 0, 0)
        t.reset()
        t.rt(90)
        t.fd(100)
        self.assertImageEquals("test_img/reset.png")

    def test_color(self):
        t = self.turtle
        for _ in range(2):
            t.fd(100)
            for _ in range(4):
                t.fd(10)
                t.rt(90)
            t.fd(100)
            t.color(255, 0, 0)
        self.assertImageEquals("test_img/test_color.png")

    def test_turtle_drawn_correctly_at_start(self):
        self.assertImageEquals("test_img/just_turtle.png", hide_turtle=False)

    def test_turtle_at_pos(self):
        t = self.turtle
        t.fd(10)
        t.rt(45)
        t.fd(20)
        self.assertImageEquals("test_img/turtle_at_pos.png", hide_turtle=False)

    def test_turtle_full_rotate(self):
        t = self.turtle
        t.fd(10)
        t.rt(45 + 360)
        t.fd(20)
        self.assertImageEquals("test_img/turtle_full_rotate.png", hide_turtle=False)

    def test_color_strings(self):
        t = self.turtle
        t.pencolor("#12fa0b")
        t.fd(100)
        t.pencolor("red")
        t.fd(100)
        t.pencolor("#1fb")
        t.fd(100)
        self.assertImageEquals("test_img/color_strings.png")

    def test_partial_circle(self):
        t = self.turtle
        t.fd(100)
        t.circle(100, 90)
        t.fd(100)
        t.circle(50, -90)
        t.fd(100)
        t.circle(-50, 90)
        t.fd(100)
        t.circle(-50, -90)
        t.fd(100)
        self.assertImageEquals("test_img/partial_circle.png")

    def test_partial_circle_fill(self):
        t = self.turtle
        t.begin_fill()
        t.fillcolor("red")
        t.fd(100)
        t.circle(100, 90)
        t.fd(100)
        t.end_fill()
        self.assertImageEquals("test_img/partial_circle_fill.png")

    def test_simple_partial_circle(self):
        t = self.turtle
        t.fd(100)
        t.circle(100, 90)
        t.fd(100)
        self.assertImageEquals("test_img/simple_partial_circle.png")

    def test_pixel(self):
        t = self.turtle
        t.pixel_size(100)
        t.pixel(0, 0, "red")
        t.pixel_size(50)
        t.pixel(1.5, 0.5, "green")
        self.assertImageEquals("test_img/test_pixel.png")

    def test_pixel_out_of_bounds(self):
        t = self.turtle
        t.pixel_size(40)
        for x in range(21):
            t.pixel(0, x, x * 10, 0, int(255 * (20 - x) / 20))
        self.assertImageEquals("test_img/test_oob_pixel.png")

    def test_1x1_pixel(self):
        self.setUp(10, 10)
        t = self.turtle
        t.pixel(0, 0, "black")
        t.pixel(1, 0, "red")
        t.pixel(1, 1, "green")
        t.pixel(0, 1, "blue")
        t.pixel(-1, 1, "yellow")
        t.pixel(-1, 0, "magenta")
        t.pixel(-1, -1, "cyan")
        t.pixel(0, -1, "orange")
        t.pixel(1, -1, "purple")
        self.assertImageEquals("test_img/test_1x1_pixel.png")

    def test_standard_mode(self):
        t = self.turtle
        t.fd(100)  # should move to the east
        t.lt(90)  # should face north
        t.fd(100)  # should move north
        t.setheading(-90)  # should face south
        t.fd(200)  # should move south
        self.assertImageEquals("test_img/test_standard_mode.png")

    def test_logo_mode(self):
        t = self.turtle
        t.mode("logo")
        t.fd(100)  # should move to the north
        t.lt(90)  # should face west
        t.fd(100)  # should move west
        t.setheading(-90)  # should stay facing west
        t.fd(200)  # should move west
        self.assertImageEquals("test_img/test_logo_mode.png")

class ForwardingTest(PillowTest):
    def setUp(self, width=1000, height=1000):
        self.underlying_canvas = PillowCanvas(width, height)
        self.canvas = ForwardingCanvas(self.underlying_canvas)
        self.turtle = Turtle(self.canvas)

    def assertImageEquals(self, path, hide_turtle=True):
        if hide_turtle:
            self.turtle.hideturtle()
        data = self.underlying_canvas.export()
        saved_data = np.array(Image.open(path).convert("RGBA"))
        self.assertTrue(np.array_equal(data, saved_data))




from abc import ABC, abstractmethod
from math import pi


class Canvas(ABC):
    def __init__(self, width, height):
        self.width = width
        self.height = height
        self._turtle = None

    @property
    def turtle(self):
        return self._turtle

    @turtle.setter
    def turtle(self, turtle):
        self._turtle = turtle
        self.refreshed_turtle(turtle)

    @abstractmethod
    def draw_rectangular_line(self, start, end, color, width):
        """
        Draw a 1 width line from START to END with the given color COLOR
        """
        pass

    def draw_line(self, start, end, color, width):
        if width > 1:
            self.draw_circle(start, width / 2, color, width, True, 0, 2 * pi)
        self.draw_rectangular_line(start, end, color, width)
        if width > 1:
            self.draw_circle(end, width / 2, color, width, True, 0, 2 * pi)

    @abstractmethod
    def draw_circle(self, center, radius, color, width, is_filled, start, end):
        """
        Draw a circle of width 1 with the given center CENTER, radius RADIUS, and color COLOR.
        Only draw the portion with angle between START and END, moving counterclockwise from START to END.
        RADIUS must be non-negative.

        Fill the circle if IS_FILLED is true.
        """
        pass

    @abstractmethod
    def fill_path(self, path, color):
        """
        Fill the given polygon formed by the movements in PATH using fill color COLOR.
        """
        pass

    @abstractmethod
    def axis_aligned_rectangle(self, bottom_left, width, height, color):
        """
        Fill the given rectangle with bottom left corner BOTTOM_LEFT, and dimensions (WIDTH, HEIGHT),
            and fill color COLOR
        """
        pass

    @abstractmethod
    def set_bgcolor(self, color):
        """
        Fill the entire background with the given COLOR
        """
        pass

    @abstractmethod
    def clear(self):
        """
        Clear everything in the foreground
        """
        pass

    @abstractmethod
    def refreshed_turtle(self, drawn_turtle):
        """
        Update the turtle to the given DrawnTurtle object, or remove the turtle if None is passed
        """
        pass

    @abstractmethod
    def set_speed(self, speed):
        """
        Set the animation speed, should be consistent with the built-in turtle module if implemented at all.
        """
        pass

    @abstractmethod
    def exit_on_click(self):
        """
        Blocks until a click, upon which the GUI is closed.
        For non-gui apps, this does nothing, but GUI based apps should change the behavior.
        """
        pass




from math import sin, cos, pi, atan2
try:
    import turtle
    import tkinter
except ImportError:
    pass

from .canvas import Canvas
from .model import Color, Position, Arc


class TkCanvas(Canvas):
    """
    Draw the given results on a pillow canvas.
    """
    def __init__(self, width, height, init_hook=lambda: None):
        super().__init__(width, height)
        self.image = None
        self.__init_hook = init_hook
        self.__init_hook_run = False

    def __ensure_init_hook(self):
        if self.__init_hook_run:
            return
        turtle.screensize(self.width, self.height)
        self.__init_hook_run = True
        self.__init_hook()

    def tr_pos(self, pos):
        x, y = pos.x, pos.y
        return round(x), round(-y)

    @staticmethod
    def tr_color(color):
        return '#%02x%02x%02x' % tuple(color)

    def _goto(self, pos):
        turtle.goto(*pos)

    def _goto_invisible(self, pos):
        speed = turtle.speed()
        try:
            turtle.speed(0)
            turtle.pu()
            self._goto(pos)
        finally:
            turtle.pd()
            turtle.speed(speed)

    def _goto_visible(self, pos, color, width):
        self.__ensure_init_hook()
        turtle.color(self.tr_color(color))
        turtle.width(width)
        self._goto(pos)

    def _setheading(self, heading):
        self.__ensure_init_hook()
        turtle.radians()
        turtle.setheading(heading)

    def draw_rectangular_line(self, start, end, color, width):
        self.__ensure_init_hook()
        self._goto_invisible(start)
        self._goto_visible(end, color, width)

    def draw_circle(self, center, radius, color, width, is_filled, start, end):
        self.__ensure_init_hook()
        if is_filled:
            assert start == 0
            assert end == 2 * pi
            x, y = self.tr_pos(center)
            left_up = (x-radius, y-radius)
            right_down = (x+radius, y+radius)
            box = [left_up, right_down]
            turtle.getcanvas().create_oval(box, fill=self.tr_color(color), width=0)
        else:
            sx, sy = center.x + radius * cos(start), center.y + radius * sin(start)
            theta = start + pi/2
            self._goto_invisible((sx, sy))
            self._setheading(theta)
            amount = end - start
            turtle.circle(radius, (end - start))

    def fill_path(self, path, color):
        self.__ensure_init_hook()
        points = []
        for movement in path:
            points += movement.to_points()
        polygon = [coor for point in points for coor in self.tr_pos(point)]
        turtle.getcanvas().create_polygon(polygon,
            fill=self.tr_color(color))
        turtle.getcanvas().update_idletasks()

    def axis_aligned_rectangle(self, bottom_left, width, height, color):
        self.__ensure_init_hook()
        x, y = bottom_left
        blx, bly = self.tr_pos(bottom_left)
        # bounding box in pixel space, make this exactly w*h
        # subtract one to have blx, blx+1, ..., blx+w-1, etc.
        tlx, tly = blx + (width - 1), bly - (height - 1)
        turtle.getcanvas().create_rectangle(blx, bly, tlx, tly, fill=self.tr_color(color), width=0)

    def set_bgcolor(self, color):
        self.__ensure_init_hook()
        turtle.bgcolor(self.tr_color(color))

    def clear(self):
        self.__ensure_init_hook()
        self.axis_aligned_rectangle(Position(-self.width / 2, -self.height / 2), self.width, self.height, Color(255, 255, 255))
        turtle.clear()

    def refreshed_turtle(self, drawn_turtle):
        self.__ensure_init_hook()
        if drawn_turtle is None:
            turtle.hideturtle()
            return
        turtle.showturtle()
        self._setheading(drawn_turtle.heading)
        self._goto_invisible(drawn_turtle.pos)
        turtle.shapesize(drawn_turtle.stretch_wid, drawn_turtle.stretch_len)

    def set_speed(self, speed):
        turtle.speed(speed)

    def export(self, path):
        """
        Exports the current image as a postscript, to the given path
        """
        turtle.getscreen().getcanvas().postscript(file=path)

    def exit_on_click(self):
        turtle.exitonclick()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="abstract_turtle", # Replace with your own username
    version="0.2.1",
    author="Kavi Gupta",
    author_email="abstract_turtle@kavigupta.org",
    description="Reimplementation of the builtin turtle module that allows for arbitrary backends.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kavigupta/abstract_turtle",
    download_url="https://github.com/kavigupta/abstract_turtle/archive/0.2.1.zip",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    install_requires=[],
    extras_require={
        "pillow_canvas" : ["numpy==1.18.0", "Pillow==6.2.1"]
    }
)



from functools import wraps
from math import pi, sin, cos, copysign

from .model import Color, Position, DrawnTurtle, Mode, LineTo, Arc
from .canvas import Canvas

def turtle_method(func):
    """
    Marks the given method as one that needs to be placed in global.
    """
    func.is_turtle_method = True
    return func


def make_formode():
    handlers = {}

    def formode(mode):
        def decorator(func):
            @wraps(func)
            def error(self, *args, **kwargs):
                raise RuntimeError(
                    "Handler not available for mode: {}".format(self._BaseTurtle__mode)
                )

            prev = handlers.get(func.__name__, error)

            @wraps(func)
            def handler(self, *args, **kwargs):
                if self._BaseTurtle__mode == mode:
                    return func(self, *args, **kwargs)
                else:
                    return prev(self, *args, **kwargs)

            handlers[func.__name__] = handler

            return handler

        return decorator

    return formode


formode = make_formode()


class BaseTurtle:
    """
    Manages all the basic turtle functionality. The other turtle methods can be expressed in terms of these.
    """
    def __init__(self, canvas):
        if not isinstance(canvas, Canvas):
            raise RuntimeError("Expected the argument to Turtle to be of type {} but was {} of type {}".format(
                Canvas.__name__,
                canvas,
                type(canvas).__name__
            ))
        self.__canvas = canvas
        self.__x = 0
        self.__y = 0
        self.__line_width = 1
        self.__theta = 0
        self.__pen_color = Color(0, 0, 0)
        self.__fill_color = Color(0, 0, 0)
        self.__pen_down = True
        self.__degrees = 360
        self.__path = None
        self.__turtle_is_shown = True
        self.__turtle_stretch_wid = 1
        self.__turtle_stretch_len = 1
        self.__pixel_size = 1
        self.__mode = Mode.STANDARD
        self.__speed = 3 # default from the normal turtle module

        self.__update_turtle()

    @turtle_method
    def goto(self, x, y):
        """
        Go to the given position (X, Y).
        """
        if self.__pen_down:
            self.__canvas.draw_line(self.__current_pos, Position(x, y), self.__pen_color, self.__line_width)
        self.__x = x
        self.__y = y
        if self.filling():
            self.__path.append(LineTo(self.__current_pos))
        self.__update_turtle()
    setpos = setposition = goto

    @turtle_method
    def forward(self, amount):
        """
        Move forward the given amount.
        """
        self.goto(self.xcor() + amount * cos(self.__theta), self.ycor() + amount * sin(self.__theta))
    fd = forward

    @turtle_method
    def setheading(self, heading):
        """
        Set the heading to the given value in degrees
        """
        self.__theta = self.__to_real_angle(heading)
        self.__update_turtle()
    seth = setheading

    @turtle_method
    def circle(self, radius, extent=None):
        """
        Draw a circle starting at the given point with the given RADIUS and EXTENT. If EXTENT exists, draw only the
        first EXTENT degrees of the circle. If RADIUS is positive, draw in the counterclockwise direction.
        Otherwise, draw in the clockwise direction.
        """
        if extent is None:
            extent = self.__degrees

        extent = extent / self.__degrees * (2 * pi)

        center = Position(
            self.__current_pos.x - radius * sin(self.__theta),
            self.__current_pos.y + radius * cos(self.__theta),
        )
        angle_change = copysign(1, radius) * extent
        start_angle = self.__theta - pi / 2 * copysign(1, radius)
        end_angle = start_angle + angle_change

        if self.filling():
            self.__path.append(Arc(center, abs(radius), start_angle, end_angle))

        if self.__pen_down:
            if radius * extent < 0:
                start_angle, end_angle = end_angle, start_angle
            self.__canvas.draw_circle(center, abs(radius), self.__pen_color, self.__line_width, False, start_angle,
                                      end_angle)

        final_pos = Position(
            center.x + radius * sin(self.__theta + angle_change),
            center.y - radius * cos(self.__theta + angle_change),
        )
        self.__theta += angle_change
        self.__x, self.__y = final_pos.x, final_pos.y
        self.__update_turtle()

    @turtle_method
    def dot(self, size=None):
        """
        Draw a dot at the current location. If size is not specified, set it to
            (pensize + 4, pensize * 2)
        """
        if size is None:
            size = max(self.__line_width + 4, self.__line_width * 2)
        if self.__pen_down:
            self.__canvas.draw_circle(self.__current_pos, size, self.__pen_color, self.__line_width, True, 0, 2 * pi)

    @turtle_method
    def pixel(self, x, y, *color):
        """
        Fill in a square of size pixel_size at (x * pixel_size, y * pixel_size) with the given color.
        """
        d = self.__pixel_size
        self.__canvas.axis_aligned_rectangle(
            Position(x * d, y * d),
            d, d,
            self.__convert_color(*color)
        )

    @turtle_method
    def pixel_size(self, pixel_size):
        if not isinstance(pixel_size, int) or pixel_size <= 0:
            raise ValueError("Expected a positive integer for pixel_size but got {}".format(pixel_size))
        self.__pixel_size = pixel_size

    @turtle_method
    def canvas_width(self):
        """
        Return the current screen size in pixel units
        """
        return self.__canvas.width // self.__pixel_size

    @turtle_method
    def canvas_height(self):
        """
        Return the current screen size in pixel units
        """
        return self.__canvas.height // self.__pixel_size

    @turtle_method
    def xcor(self):
        """
        Get the current x coordinate
        """
        return self.__x

    @turtle_method
    def ycor(self):
        """
        Get the current y coordinate
        """
        return self.__y

    @turtle_method
    def heading(self):
        """
        Get the current heading
        """
        return self.__from_real_angle(self.__theta)

    @turtle_method
    def degrees(self, amount=360):
        """
        Set the number of degrees in a circle
        """
        self.__degrees = amount

    @turtle_method
    def pendown(self):
        """
        Do draw when moving
        """
        self.__pen_down = True
    pd = down = pendown

    @turtle_method
    def penup(self):
        """
        Do not draw when moving
        """
        self.__pen_down = False
    pu = up = penup

    @turtle_method
    def pensize(self, width=None):
        """
        Set or get the pen size. If WIDTH is None, get it, otherwise set it.
        """
        if width is None:
            return self.__line_width
        self.__line_width = width
    width = pensize

    @turtle_method
    def isdown(self):
        """
        Return if the pen is down or not
        """
        return self.__pen_down

    @turtle_method
    def pencolor(self, *color):
        """
        Set the pen color as COLOR
        """
        self.__pen_color = self.__convert_color(*color)

    @turtle_method
    def fillcolor(self, *color):
        """
        Set the fill color as COLOR
        """
        self.__fill_color = self.__convert_color(*color)

    @turtle_method
    def filling(self):
        """
        Return whether the canvas is filling.
        """
        return self.__path is not None

    @turtle_method
    def begin_fill(self):
        """
        Begin setting the polygon to fill
        """
        self.__path = [LineTo(self.__current_pos)]

    @turtle_method
    def end_fill(self):
        """
        End setting the polygon to fill, and fill it in.
        """
        if self.__path is None:
            return
        self.__canvas.fill_path(self.__path, self.__fill_color)
        self.__path = None

    @turtle_method
    def clear(self):
        """
        Clear the canvas, but do not move the turtle.
        """
        self.__canvas.clear()

    @turtle_method
    def bgcolor(self, *color):
        self.__canvas.set_bgcolor(self.__convert_color(*color))

    def __update_turtle(self):
        if self.__turtle_is_shown:
            self.__canvas.turtle = DrawnTurtle(self.__current_pos, self.__theta % (2 * pi), self.__turtle_stretch_wid, self.__turtle_stretch_len)
        else:
            self.__canvas.turtle = None

    @turtle_method
    def hideturtle(self):
        """
        Hide the turtle from the canvas.
        """
        self.__turtle_is_shown = False
        self.__update_turtle()
    ht = hideturtle

    @turtle_method
    def showturtle(self):
        """
        Show the turtle on the canvas
        """
        self.__turtle_is_shown = True
        self.__update_turtle()
    st = showturtle

    @turtle_method
    def isvisible(self):
        """
        Return whether the turtle is visible
        """
        return self.__turtle_is_shown

    @turtle_method
    def shapesize(self, stretch_wid=None, stretch_len=None):
        self.__turtle_stretch_wid = stretch_wid
        self.__turtle_stretch_len = stretch_len
        self.__update_turtle()
    turtlesize = shapesize

    @turtle_method
    def mode(self, mode=None):
        if mode is None:
            return self.__mode.value
        elif mode == "standard":
            self.__mode = Mode.STANDARD
        elif mode == "logo":
            self.__mode = Mode.LOGO
        elif mode == "world":
            raise RuntimeError("Custom world coordinates not supported.")
        else:
            raise RuntimeError("Unknown mode: {}".format(mode))
        self.goto(0, 0)
        self.setheading(0)
        self.clear()

    @turtle_method
    def speed(self, speed=None):
        if speed is None:
            return self.__speed
        self.__speed = speed
        self.__canvas.set_speed(speed)

    @turtle_method
    def exitonclick(self):
        return self.__canvas.exit_on_click()

    @property
    def __current_pos(self):
        return Position(self.__x, self.__y)

    @formode(Mode.STANDARD)
    def __to_real_angle(self, amount):
        return (amount / self.__degrees) * (2 * pi)

    @formode(Mode.STANDARD)
    def __from_real_angle(self, angle):
        return (angle / (2 * pi)) * self.__degrees % self.__degrees

    @formode(Mode.LOGO)
    def __to_real_angle(self, amount):
        return (1 / 4 - amount / self.__degrees) * (2 * pi)

    @formode(Mode.LOGO)
    def __from_real_angle(self, angle):
        return (1 / 4 - angle / (2 * pi)) * self.__degrees % self.__degrees

    @staticmethod
    def __convert_color(*color):
        return Color.of(*color)


class Turtle(BaseTurtle):
    """
    This entire class should only use public methods of the BaseTurtle class.
    """

    @turtle_method
    def backward(self, amount):
        """
        Move backward the given amount.
        """
        self.forward(-amount)
    bk = back = backward

    @formode(Mode.STANDARD)
    def right(self, amount):
        self.setheading(self.heading() - amount)

    @turtle_method
    @formode(Mode.LOGO)
    def right(self, amount):
        """
        Rotate right the given amount.
        """
        self.setheading(self.heading() + amount)

    rt = right

    @turtle_method
    def left(self, amount):
        """
        Rotate left the given amount.
        """
        self.right(-amount)
    lt = left

    @turtle_method
    def setx(self, x):
        """
        Move so that the x coordinate is X
        """
        self.goto(x, self.xcor())

    @turtle_method
    def sety(self, y):
        """
        Move so that the y coordinate is Y
        """
        self.goto(self.xcor(), y)

    @turtle_method
    def home(self):
        """
        Set location to (0, 0) and set heading to 0
        """
        self.goto(0, 0)
        self.setheading(0)

    @turtle_method
    def position(self):
        """
        Get the current position as a tuple
        """
        return self.xcor(), self.ycor()
    pos = position

    @turtle_method
    def distance(self, other):
        """
        Get the distance between this and the other location/turtle.
        """
        if isinstance(other, Turtle):
            return self.distance(other.position())
        x, y = other
        return ((x - self.xcor()) ** 2 + (y - self.ycor()) ** 2) ** 0.5

    @turtle_method
    def radians(self):
        """
        Set angle units to radians
        """
        return self.degrees(2 * pi)

    @turtle_method
    def color(self, *color):
        """
        Set both the pen and fill colors
        """
        self.pencolor(*color)
        self.fillcolor(*color)

    @turtle_method
    def reset(self):
        self.home()
        self.clear()
